/**
 * Cecile Janin ad9953 driver
 */

#include <app/drivers/ad9953.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/drivers/spi.h>
#include <zephyr/logging/log.h>
#include <zephyr/init.h>
#include <zephyr/kernel.h>
#include <zephyr/logging/log.h>

LOG_MODULE_REGISTER(ad9953, CONFIG_AD9953_LOG_LEVEL);

struct ad9953_config {
	const uint16_t clk_frequency;

	const struct gpio_dt_spec io_update_gpios;
	const struct gpio_dt_spec pwr_dwn_ctrl_gpios;
	const struct gpio_dt_spec osk_gpios;
	const struct gpio_dt_spec ps0_gpios;
	const struct gpio_dt_spec ps1_gpios;

	const struct spi_dt_spec spi;
};

struct ad9953_data {
	uint8_t clock_multiplier;
};

#define IO_UPDATE_DELAY K_USEC(10)
static inline void io_update_toggle(const struct device *dev) {
	const struct ad9953_config *config = dev->config;
	gpio_pin_set_dt(&config->io_update_gpios, 1);
	k_sleep(IO_UPDATE_DELAY);
	gpio_pin_set_dt(&config->io_update_gpios, 0);
}

/**
 * @brief Write ad9953 register
 * 
 * Write data in the given register
 * 
 * @param dev Pointer to the device
 * @param reg to write
 * @param reg_size the register size
 * @param data_to_send data to write
 * 
 * @retval 0 on success;
 * @retval <0 negative errno code on error. 
 */
static int ad9953_write(const struct device *dev,  uint8_t reg, 
	uint8_t length, uint8_t * data_to_send) {

	const struct ad9953_config *config = dev->config;
	int ret = 0;
	uint8_t data[length+1];

	data[0] = reg;
	for (int i = 0; i < length; ++i)
	{
		data[i+1] = data_to_send[i];
	}

	io_update_toggle(dev);

	struct spi_buf tx_buf = {
		.buf = data,
		.len = length+1
	};

	struct spi_buf_set tx_bufs = {
		.buffers = &tx_buf,
		.count = 1
	};

	ret = spi_write_dt(&config->spi, &tx_bufs);

	io_update_toggle(dev);

	return ret;
}

/**
 * @brief Read ad9953 register
 * 
 * Read the given register
 * 
 * @param dev Pointer to the device
 * @param reg to read
 * @param reg_size the register size
 * @param read_buffer to store read data
 * 
 * @retval 0 on success;
 * @retval <0 negative errno code on error. 
 */
int ad9953_read(const struct device *dev, uint8_t reg, 
	uint8_t reg_size, uint8_t * read_buffer) {

	int ret = 0;
	const struct ad9953_config *config = dev->config;
	uint8_t data_to_send;

	if ( (read_buffer == NULL) | (dev == NULL) | (reg_size > 4) )
	{
		return -EINVAL;
	}

	/* Send the Read command */
	data_to_send = BIT(7) | reg; /* read bit + bit register */
	LOG_WRN("read byte : 0b%x", data_to_send);

	struct spi_buf tx_buf = {
		.buf = &data_to_send,
		.len = 1
	};

	struct spi_buf_set tx_bufs = {
		.buffers = &tx_buf,
		.count = 1
	};

	ret = spi_write_dt(&config->spi, &tx_bufs);

	tx_buf.len = reg_size;
	tx_buf.buf = read_buffer;
	spi_read_dt(&config->spi, &tx_bufs);

	io_update_toggle(dev);

	return ret;
}

int ad9953_set_freq_tuning_word(const struct device *dev, uint32_t frequency_out){
	const struct ad9953_config *config = dev->config;
	struct ad9953_data *data = dev->data;
	int ret = 0;
	float frequency_word = 0;
	uint8_t send_buffer[4];

	if (dev == NULL)
	{
		LOG_ERR("Bad device : %s", dev->name);
		return -EINVAL;
	} else if (frequency_out == 0)
	{
		for (int i = 0; i < 4; ++i)
		{
			send_buffer[i] = 0;
		}
		LOG_WRN("Disabling generator : %s", dev->name);
		ret = ad9953_write(dev, FREQUENCY_TUNING_WORD_0, 
		FREQUENCY_TUNING_WORD_0_SIZE, send_buffer);
	} 
	else if ((frequency_out < 100000) | (frequency_out > 20000000))
	{
		LOG_ERR("freq val out of range : %d", frequency_out);
		return -EINVAL;
	}

	frequency_word =  data->clock_multiplier * config->clk_frequency;
	frequency_word *= 1000000;
	frequency_word = frequency_out / frequency_word;
	frequency_word *= 4294967296;

	uint32_t test_val = frequency_word;

	for (uint8_t i = 0; i < 4; ++i)
	{
		send_buffer[i] = test_val >> (3-i)*8;
		//LOG_WRN("freq : %x, %d", send_buffer[i], i);
	}

	ret = ad9953_write(dev, FREQUENCY_TUNING_WORD_0, 
		FREQUENCY_TUNING_WORD_0_SIZE, send_buffer);

    return ret;
}

int ad9953_set_pll(const struct device *dev, uint8_t multiplier)
{
	uint8_t data_to_send[3];
	int ret = 0;

	if (dev == NULL)
	{
		LOG_ERR("Bad device : %s", dev->name);
		return -EINVAL;
	}
	else if (multiplier > 20)
	{
		LOG_ERR("PLL val out of range : %d", multiplier);
		return -EINVAL;
	}

	multiplier = multiplier << 3;


	data_to_send[0] = 0x18;
	data_to_send[1] = 0x00;
	data_to_send[2] = multiplier; 
	ret = ad9953_write(dev, CONTROL_FUNCTION_REGISTER_2, 
		CONTROL_FUNCTION_REGISTER_2_SIZE, data_to_send);

	return ret;
}

int ad9953_disable(const struct device *dev)
{
	int ret =0;
	uint8_t send_buffer[4];

	if (dev == NULL)
	{
		LOG_ERR("Bad device : %s", dev->name);
		return -EINVAL;
	}

	for (int i = 0; i < 4; ++i)
	{
		send_buffer[i] = 0;
	}

	ret = ad9953_write(dev, FREQUENCY_TUNING_WORD_0, 
		FREQUENCY_TUNING_WORD_0_SIZE, send_buffer);

	return ret;
}

static int ad9953_init(const struct device *dev)
{
	int ret = 0;
	const struct ad9953_config *config = dev->config;
	struct ad9953_data *data = dev->data;

	if (!spi_is_ready_dt(&config->spi))
	{
		LOG_ERR("spi bus not ready");
		return -ENODEV;
	}
	
	ret = gpio_is_ready_dt(&config->io_update_gpios);
	ret = gpio_pin_configure_dt(&config->io_update_gpios, GPIO_OUTPUT);
	ret = gpio_pin_set_dt(&config->io_update_gpios, 0);

	ret = gpio_is_ready_dt(&config->pwr_dwn_ctrl_gpios);
	ret = gpio_pin_configure_dt(&config->pwr_dwn_ctrl_gpios, GPIO_OUTPUT);
	ret = gpio_pin_set_dt(&config->pwr_dwn_ctrl_gpios, 0);

	ret = gpio_is_ready_dt(&config->osk_gpios);
	ret = gpio_pin_configure_dt(&config->osk_gpios, GPIO_OUTPUT);
	ret = gpio_pin_set_dt(&config->osk_gpios, 0);

	ret = gpio_is_ready_dt(&config->ps0_gpios);
	ret = gpio_pin_configure_dt(&config->ps0_gpios, GPIO_OUTPUT);
	ret = gpio_pin_set_dt(&config->ps0_gpios, 0);

	ret = gpio_is_ready_dt(&config->ps1_gpios);
	ret = gpio_pin_configure_dt(&config->ps1_gpios, GPIO_OUTPUT);
	ret = gpio_pin_set_dt(&config->ps1_gpios, 0);

	if (ret)
	{
		LOG_ERR("gpio conf error : %x", ret);
	}

	/* debut test */

	uint8_t data_to_send[5];

	data_to_send[0] = 0x00;
	data_to_send[1] = 0x00;
	data_to_send[2] = AUTOCLEAR_PHASE_ACC_bm | SDIO_INPUT_ONLY_bm; 
	data_to_send[3] = 0x00;

	ad9953_write(dev, CONTROL_FUNCTION_REGISTER_1, 
		CONTROL_FUNCTION_REGISTER_1_SIZE, data_to_send);

	
	data_to_send[0] = 0x18;
	data_to_send[1] = 0x00;
	data_to_send[2] = REFCLK_MULITPLIER_8X_bm; 
	ad9953_write(dev, CONTROL_FUNCTION_REGISTER_2, 
		CONTROL_FUNCTION_REGISTER_2_SIZE, data_to_send);

	data->clock_multiplier = 8;

	data_to_send[0] = 0x00;
	data_to_send[1] = 0x00;
	ad9953_write(dev, AMPLITUDE_SCALE_FACTOR, 
		AMPLITUDE_SCALE_FACTOR_SIZE, data_to_send);


	data_to_send[0] = 0x00;
	ad9953_write(dev, AMPLITUDE_RAMP_RATE, 
		AMPLITUDE_RAMP_RATE_SIZE, data_to_send);

	//ad9953_set_freq_tuning_word(dev, 0);

	return ret;
}

#define AD9953_DEFINE(node_id)                               \
	static const struct ad9953_config config##node_id = {    \
		.io_update_gpios = GPIO_DT_SPEC_GET_OR(node_id, io_update_gpios, {0}),  \
		.pwr_dwn_ctrl_gpios = GPIO_DT_SPEC_GET_OR(node_id, pwr_dwn_ctrl_gpios, {0}),  \
		.osk_gpios = GPIO_DT_SPEC_GET_OR(node_id, osk_gpios, {0}),  \
		.ps0_gpios = GPIO_DT_SPEC_GET_OR(node_id, ps0_gpios, {0}),  \
		.ps1_gpios = GPIO_DT_SPEC_GET_OR(node_id, ps1_gpios, {0}),  \
		.clk_frequency = DT_PROP(node_id, clk_frequency),  \
		.spi = SPI_DT_SPEC_GET(node_id, SPI_OP_MODE_MASTER | \
			SPI_TRANSFER_MSB | SPI_WORD_SET(8), 0),	\
	};                                                          \
	static struct ad9953_data data##node_id = {    \
		.clock_multiplier = 1,	\
	};	\
	\
	DEVICE_DT_DEFINE(node_id, ad9953_init, NULL, &data##node_id, \
		&config##node_id, POST_KERNEL, CONFIG_AD9953_INIT_PRIORITY, \
		NULL);

DT_FOREACH_STATUS_OKAY(adi_ad9953, AD9953_DEFINE)
