#ifndef DRIVER_AD9953_
#define DRIVER_AD9953_

#include <zephyr/kernel.h>

enum REGISTERS
{
    CONTROL_FUNCTION_REGISTER_1 = 0, /* CFR1 */
    CONTROL_FUNCTION_REGISTER_2,    /* CFR2 */
    AMPLITUDE_SCALE_FACTOR, /* ASF */
    AMPLITUDE_RAMP_RATE, /* ARR */
    FREQUENCY_TUNING_WORD_0, /* FTW0 */
    PHASE_OFFSET_WORD, /* POW0 */
    FREQUENCY_TUNING_WORD_1, /* FTW1 */
    RAM_SEGMENT_CONTROL_0, /* RSCW0 */
    RAM_SEGMENT_CONTROL_1, /* RSCW1 */
    RAM_SEGMENT_CONTROL_2, /* RSCW2 */
    RAM_SEGMENT_CONTROL_3, /* RSCW3 */
    RAM
};

enum REGISTERS_SIZE
{
    CONTROL_FUNCTION_REGISTER_1_SIZE = 4, /* CFR1 */
    CONTROL_FUNCTION_REGISTER_2_SIZE = 3,    /* CFR2 */
    AMPLITUDE_SCALE_FACTOR_SIZE = 2, /* ASF */
    AMPLITUDE_RAMP_RATE_SIZE = 1, /* ARR */
    FREQUENCY_TUNING_WORD_0_SIZE = 4, /* FTW0 */
    PHASE_OFFSET_WORD_SIZE = 2, /* POW0 */
    FREQUENCY_TUNING_WORD_1_SIZE = 4, /* FTW1 */
    RAM_SEGMENT_CONTROL_0_SIZE = 5, /* RSCW0 */
    RAM_SEGMENT_CONTROL_1_SIZE = 5, /* RSCW1 */
    RAM_SEGMENT_CONTROL_2_SIZE = 5, /* RSCW2 */
    RAM_SEGMENT_CONTROL_3_SIZE = 5, /* RSCW3 */
    RAM_SIZE = 128
};

/* Bitmask for CFR1 */
/* Byte 1 7:0 */
#define SYNK_CLK_DIS_bm BIT(1)
#define LINEAR_SWEEP_NO_DWELL_bm BIT(2)
#define EXTERNAL_POWER_DOWN_MODE_bm BIT(3)
#define CLOCK_INPUT_POWER_DOWN_bm BIT(4)
#define DAC_POWER_DOWN_bm BIT(5)
#define COMP_POWER_DOWN_bm BIT(6)
#define DIGITAL_POWER_DOWN_bm BIT(7)
/* Byte 2 15:8 */
#define LSB_FIRST_bm BIT(0)
#define SDIO_INPUT_ONLY_bm BIT(1)
#define CLEAR_PHASE_ACC_bm BIT(2)
#define CLEAR_FREQUENCY_ACC_bm BIT(3)
#define ENABLE_SIN_OUTPUT_bm BIT(4)
#define AUTOCLEAR_PHASE_ACC_bm BIT(5)
#define AUTOCLEAR_FREQUENCY_ACC_bm BIT(6)
#define LOAD_SRR_AT_IO_bm BIT(7)
/* Byte 3 23:16 */
#define LINEAR_SWEEP_ENABLE_bm BIT(5)
#define SOFTWARE_MANUAL_SYNC_bm BIT(6)
#define AUTOMATIC_SYNC_ENABLE_bm BIT(7)
/* Byte 4 31:24 */
#define AUTO_OSK_KEYING_bm BIT(0)
#define OSK_ENABLE_bm BIT(1)
#define LOAD_ARR_AT_IO_bm BIT(2)
#define RAM_DEST_IS_PHASE_WORD_bm BIT(6)
#define RAM_ENABLE_bm BIT(7)

/* Bitmask for CFR2 */
/* Byte 1 7:0 */
enum CHARGE_PUMP_CURRENT /* bits 0 & 1 */ 
{
    CHARGE_PUMP_CURRENT_75UA_bm = 0,
    CHARGE_PUMP_CURRENT_100UA_bm,
    CHARGE_PUMP_CURRENT_125UA_bm,
    CHARGE_PUMP_CURRENT_150UA_bm
};
#define VCO_RANGE BIT(2);
enum REFCLK_MULTIPLIER /* bits 3 to 7 */ 
{
    REFCLK_MULITPLIER_0X_bm = 0,
    REFCLK_MULITPLIER_4X_bm = 0x20,
    REFCLK_MULITPLIER_5X_bm = 0x28,
    REFCLK_MULITPLIER_6X_bm = 0x30,
    REFCLK_MULITPLIER_7X_bm = 0x38,
    REFCLK_MULITPLIER_8X_bm = 0x40,
    REFCLK_MULITPLIER_9X_bm = 0x48,
    REFCLK_MULITPLIER_10X_bm = 0x50,
    REFCLK_MULITPLIER_11X_bm = 0x58,
    REFCLK_MULITPLIER_12X_bm = 0x60,
    REFCLK_MULITPLIER_13X_bm = 0x68,
    REFCLK_MULITPLIER_14X_bm = 0x70,
    REFCLK_MULITPLIER_15X_bm = 0x78,
    REFCLK_MULITPLIER_16X_bm = 0x80,
    REFCLK_MULITPLIER_17X_bm = 0x88,
    REFCLK_MULITPLIER_18X_bm = 0x90,
    REFCLK_MULITPLIER_19X_bm = 0x98,
    REFCLK_MULITPLIER_20X_bm = 0xa0
};
/* Byte 1 15:8 */
#define CRYSTAL_OUT_PIN_bm BIT(1)
#define HARDWARE_MANUAL_SYNC_ENABLE_bm BIT(2)
#define HIGH_SPEED_SYNC_ENABLE_bm BIT(3)

/**
 * @brief Set an output frequency
 * 
 * Calculate the frequency tuning word from
 * the given frequency and send it to the ad9953
 * 
 * @param dev Pointer to the device
 * @param frequency_out in Hz
 * 
 * @retval 0 on success;
 * @retval <0 negative errno code on error. 
 */
int ad9953_set_freq_tuning_word(const struct device *dev, 
        uint32_t frequency_out);

/**
 * @brief Set the internal PLL
 * 
 * Change the internal PLL to the given
 * ratio
 * 
 * @param dev Pointer to the device
 * @param multiplier value, must be 0 or 4 to 20
 * 
 * @retval 0 on success;
 * @retval <0 negative errno code on error. 
 */
int ad9953_set_pll(const struct device *dev, uint8_t multiplier);

/**
 * @brief Disable the generator
 * 
 * Disable the output signal
 * 
 * @param dev Pointer to the device
 * 
 * @retval 0 on success;
 * @retval <0 negative errno code on error. 
 */
int ad9953_disable(const struct device *dev);

#endif //DRIVER_AD9953_